import http.client
from unittest import mock
import unittest

import responses

from cumulusci.tasks.release_notes.exceptions import GithubIssuesError
from cumulusci.tasks.release_notes.generator import GithubReleaseNotesGenerator
from cumulusci.tasks.release_notes.parser import BaseChangeNotesParser
from cumulusci.tasks.release_notes.parser import ChangeNotesLinesParser
from cumulusci.tasks.release_notes.parser import GithubIssuesParser
from cumulusci.tasks.release_notes.parser import GithubLinesParser
from cumulusci.tasks.release_notes.parser import IssuesParser
from cumulusci.tasks.release_notes.parser import InstallLinkParser
from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.core.github import get_github_api
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.tests.utils import MockUtil

PARSER_CONFIG = [
    {
        "class_path": "cumulusci.tasks.release_notes.parser.GithubLinesParser",
        "title": "Critical Changes",
    },
    {
        "class_path": "cumulusci.tasks.release_notes.parser.GithubLinesParser",
        "title": "Changes",
    },
    {
        "class_path": "cumulusci.tasks.release_notes.parser.GithubIssuesParser",
        "title": "Issues Closed",
    },
]


class TestBaseChangeNotesParser(unittest.TestCase):
    def test_parse(self):
        parser = BaseChangeNotesParser("Title")
        with self.assertRaises(NotImplementedError):
            parser.parse()

    def test_render(self):
        parser = BaseChangeNotesParser("Title")
        with self.assertRaises(NotImplementedError):
            parser.render()


class TestChangeNotesLinesParser(unittest.TestCase):
    def setUp(self):
        self.title = "Title"

    def test_parse_no_start_line(self):
        change_note = "foo\r\nbar\r\n"
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, [])
        self.assertFalse(line_added)

    def test_parse_start_line_no_content(self):
        change_note = "# {}\r\n\r\n".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, [])
        self.assertFalse(line_added)

    def test_parse_start_line_no_end_line(self):
        change_note = "# {}\r\nfoo\r\nbar".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, ["foo", "bar"])
        self.assertEqual(True, line_added)

    def test_parse_start_line_end_at_header(self):
        change_note = "# {}\r\nfoo\r\n# Another Header\r\nbar".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, ["foo"])
        self.assertTrue(line_added)

    def test_parse_start_line_no_content_no_end_line(self):
        change_note = "# {}".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, [])
        self.assertFalse(line_added)

    def test_parse_multiple_start_lines_without_end_lines(self):
        change_note = "# {0}\r\nfoo\r\n# {0}\r\nbar\r\n".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, ["foo", "bar"])
        self.assertTrue(line_added)

    def test_parse_multiple_start_lines_with_end_lines(self):
        change_note = "# {0}\r\nfoo\r\n\r\n# {0}\r\nbar\r\n\r\nincluded\r\n\r\n# not included".format(
            self.title
        )
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, ["foo", "bar", "included"])
        self.assertTrue(line_added)

    def test_parse_multi_level_indent(self):
        change_note = "# {0}\r\nfoo \r\n    bar  \r\n        baz \r\n".format(
            self.title
        )
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(parser.content, ["foo", "    bar", "        baz"])
        self.assertTrue(line_added)

    def test_parse_subheading(self):
        change_note = "# {0}\r\n## Subheading\r\nfoo".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual([], parser.content)
        self.assertEqual({"Subheading": ["foo"]}, parser.h2)
        self.assertTrue(line_added)

    def test_parse_subheading_from_another_section(self):
        change_note = "## Subheading\r\n# {0}\r\nfoo".format(self.title)
        parser = ChangeNotesLinesParser(None, self.title)
        line_added = parser.parse(change_note)
        self.assertEqual(["foo"], parser.content)
        self.assertEqual({}, parser.h2)
        self.assertTrue(line_added)

    def test_render_no_content(self):
        parser = ChangeNotesLinesParser(None, self.title)
        self.assertEqual(parser.render(), "")

    def test_render_one_content(self):
        parser = ChangeNotesLinesParser(None, self.title)
        content = ["foo"]
        parser.content = content
        self.assertEqual(
            parser.render(), "# {}\r\n\r\n{}".format(self.title, content[0])
        )

    def test_render_multiple_content(self):
        parser = ChangeNotesLinesParser(None, self.title)
        content = ["foo", "bar"]
        parser.content = content
        self.assertEqual(
            parser.render(), "# {}\r\n\r\n{}".format(self.title, "\r\n".join(content))
        )

    def test_render_subheadings(self):
        parser = ChangeNotesLinesParser(None, self.title)
        parser.h2 = {"Subheading": ["foo"]}
        self.assertEqual(
            parser.render(),
            "# {}\r\n\r\n\r\n## Subheading\r\n\r\nfoo".format(self.title),
        )


class TestGithubLinesParser(unittest.TestCase):
    def setUp(self):
        self.title = "Title"

    def test_parse(self):
        generator = mock.Mock(link_pr=True)
        parser = GithubLinesParser(generator, self.title)
        pr = mock.Mock(
            number=1, html_url="http://pr", body="# {}\r\n\r\nfoo".format(self.title)
        )
        parser.parse(pr)
        self.assertEqual(1, parser.pr_number)
        self.assertEqual("http://pr", parser.pr_url)
        self.assertEqual(["foo [[PR1](http://pr)]"], parser.content)

    def test_parse_empty_pull_request_body(self):
        generator = mock.Mock(link_pr=True)
        parser = GithubLinesParser(generator, self.title)
        pr = mock.Mock(number=1, html_url="http://pr", body=None)
        line_added = parser.parse(pr)
        assert not line_added


class TestIssuesParser(unittest.TestCase):
    def setUp(self):
        self.title = "Issues"

    def test_issue_numbers(self):
        change_note = "# {}\r\nfix #2\r\nfix #3\r\nfix #5\r\n".format(self.title)
        parser = IssuesParser(None, self.title)
        parser.parse(change_note)
        self.assertEqual(parser.content, [2, 3, 5])

    def test_issue_numbers_with_links(self):
        change_note = "# {}\r\nfix [#2](https://issue)\r\nfix [#3](http://issue)\r\nfix #5\r\n".format(
            self.title
        )
        parser = IssuesParser(None, self.title)
        parser.parse(change_note)
        self.assertEqual(parser.content, [2, 3, 5])

    def test_issue_numbers_and_other_numbers(self):
        change_note = "# {}\r\nfixes #2 but not # 3 or 5".format(self.title)
        parser = IssuesParser(None, self.title)
        parser.parse(change_note)
        self.assertEqual(parser.content, [2])

    def test_multiple_issue_numbers_per_line(self):
        change_note = "# {}\r\nfix #2 also does fix #3 and fix #5\r\n".format(
            self.title
        )
        parser = IssuesParser(None, self.title)
        parser.parse(change_note)
        self.assertEqual(parser.content, [2, 3, 5])

    def test_render(self):
        parser = IssuesParser(None, self.title)
        parser.content = ["1: foo"]
        self.assertEqual("# Issues\r\n\r\n#1: foo", parser.render())


class TestGithubIssuesParser(unittest.TestCase, GithubApiTestMixin):
    def setUp(self):
        self.init_github()
        self.gh = get_github_api("TestUser", "TestPass")
        self.title = "Issues"
        # Set up the mock release_tag lookup response
        self.issue_number_valid = 123
        self.issue_number_invalid = 456
        self.pr_number = 789
        self.pr_url = "https://github.com/{}/{}/pulls/{}".format(
            "TestOwner", "TestRepo", self.pr_number
        )
        self.mock_util = MockUtil("TestOwner", "TestRepo")

    @responses.activate
    def test_issue_numbers(self):
        self.mock_util.mock_get_repo()
        change_note = "# {}\r\nFixes #2, Closed #3 and Resolve #5".format(self.title)
        self.mock_util.mock_pull_request(self.pr_number, body=change_note)
        generator = self._create_generator()
        repo = generator.get_repo()
        pull_request = repo.pull_request(self.pr_number)
        parser = GithubIssuesParser(generator, self.title)
        parser.parse(pull_request)
        pr_url = "https://github.com/TestOwner/TestRepo/pulls/{}".format(self.pr_number)
        expected_content = self._create_expected_content([2, 3, 5], pr_url)
        self.assertEqual(parser.content, expected_content)

    @responses.activate
    def test_issue_numbers_and_other_numbers(self):
        self.mock_util.mock_get_repo()
        change_note = "# {}\r\nFixes #2 but not #5".format(self.title)
        self.mock_util.mock_pull_request(self.pr_number, body=change_note)
        generator = self._create_generator()
        repo = generator.get_repo()
        pull_request = repo.pull_request(self.pr_number)
        parser = GithubIssuesParser(generator, self.title)
        parser.parse(pull_request)
        pr_url = "https://github.com/TestOwner/TestRepo/pulls/{}".format(self.pr_number)
        expected_content = self._create_expected_content([2], pr_url)
        self.assertEqual(parser.content, expected_content)

    @responses.activate
    def test_no_issue_numbers(self):
        pr_number = 1
        self.mock_util.mock_get_repo()
        change_note = "# {}\r\n#2 and #3 are fixed by this change".format(self.title)
        self.mock_util.mock_pull_request(pr_number, body=change_note)
        generator = self._create_generator()
        repo = generator.get_repo()
        pull_request = repo.pull_request(pr_number)
        parser = GithubIssuesParser(generator, self.title)
        parser.parse(pull_request)
        self.assertEqual(parser.content, [])

    @responses.activate
    def test_render_issue_number_valid(self):
        api_url = "{}/issues/{}".format(self.repo_api_url, self.issue_number_valid)
        expected_response = self._get_expected_issue(self.issue_number_valid)
        self.mock_util.mock_get_repo()
        responses.add(method=responses.GET, url=api_url, json=expected_response)
        generator = self._create_generator()
        generator.link_pr = True
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": self.issue_number_valid,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        expected_render = self._create_expected_render(
            self.issue_number_valid, expected_response["title"], True
        )
        self.assertEqual(parser.render(), expected_render)

    @responses.activate
    def test_render_issue_number_invalid(self):
        api_url = "{}/issues/{}".format(self.repo_api_url, self.issue_number_invalid)
        expected_response = self._get_expected_not_found()
        self.mock_util.mock_get_repo()
        responses.add(
            method=responses.GET,
            url=api_url,
            json=expected_response,
            status=http.client.NOT_FOUND,
        )
        generator = self._create_generator()
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": self.issue_number_invalid,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        with self.assertRaises(GithubApiNotFoundError):
            parser.render()

    def test_init__issues_disabled(self):
        generator = mock.Mock(has_issues=False)
        with self.assertRaises(GithubIssuesError):
            GithubIssuesParser(generator, self.title)

    def _create_expected_content(self, issue_numbers, pr_url):
        y = []
        for n in issue_numbers:
            y.append({"issue_number": n, "pr_number": self.pr_number, "pr_url": pr_url})
        return y

    def _create_expected_render(self, issue_number, issue_title, link_pr):
        render = "# {}\r\n\r\n#{}: {}".format(self.title, issue_number, issue_title)
        if link_pr:
            render += " [[PR{}]({})]".format(self.pr_number, self.pr_url)
        return render

    def _create_generator(self):
        generator = GithubReleaseNotesGenerator(
            self.gh, self.github_info.copy(), PARSER_CONFIG, "release/1.1"
        )
        return generator


class TestCommentingGithubIssuesParser(unittest.TestCase, GithubApiTestMixin):
    def setUp(self):
        self.init_github()
        self.gh = get_github_api("TestUser", "TestPass")
        self.mock_util = MockUtil("TestOwner", "TestRepo")
        self.title = "Issues"
        self.issue_number_without_comments = 1
        self.issue_number_with_beta_comment = 2
        self.issue_number_without_beta_comment = 3
        self.issue_number_with_prod_comment = 4
        self.issue_number_without_prod_comment = 5
        self.pr_number = 6
        self.pr_url = "https://github.com/TestOwner/TestRepo/pulls/{}".format(
            self.pr_number
        )
        self.tag_prod = "release/1.2"
        self.tag_beta = "beta/1.2-Beta_3"
        self.tag_not_prod_or_beta = "foo"
        self.version_number_prod = "1.1"
        self.version_number_beta = "1.2 (Beta 3)"

    def _create_generator(self, tag):
        generator = GithubReleaseNotesGenerator(
            self.gh, self.github_info.copy(), PARSER_CONFIG, tag, publish=True
        )
        return generator

    @responses.activate
    def test_render_issue_without_comments(self):
        issue_number = self.issue_number_without_comments
        tag = self.tag_not_prod_or_beta
        self.mock_util.mock_get_repo()
        self.mock_util.mock_post_comment(issue_number)

        # Mock the issue
        api_url = "{}/issues/{}".format(self.repo_api_url, issue_number)
        expected_issue = self._get_expected_issue(issue_number)
        responses.add(method=responses.GET, url=api_url, json=expected_issue)

        # Mock the comments list
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        responses.add(
            method=responses.GET, url=api_url, json=[], content_type="application/json"
        )

        generator = self._create_generator(tag)
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": issue_number,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        expected_render = self._create_expected_render(
            issue_number, expected_issue["title"], False
        )
        render = parser.render()
        self.assertEqual(render, expected_render)
        self.assertEqual(len(responses.calls._calls), 2)

    @responses.activate
    def test_render_issue_with_beta_comment(self):
        issue_number = self.issue_number_with_beta_comment
        tag = self.tag_beta
        self.mock_util.mock_get_repo()
        self.mock_util.mock_post_comment(issue_number)

        # Mock the issue
        api_url = "{}/issues/{}".format(self.repo_api_url, issue_number)
        expected_issue = self._get_expected_issue(issue_number)
        responses.add(method=responses.GET, url=api_url, json=expected_issue)

        # Mock the comments list
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        expected_comment_1 = self._get_expected_issue_comment(
            GithubIssuesParser.ISSUE_COMMENT["beta"]
        )
        expected_comments = [expected_comment_1]
        responses.add(method=responses.GET, url=api_url, json=expected_comments)

        generator = self._create_generator(tag)
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": issue_number,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        expected_render = self._create_expected_render(
            issue_number, expected_issue["title"], False
        )
        render = parser.render()
        self.assertEqual(render, expected_render)
        self.assertEqual(len(responses.calls._calls), 3)

    @responses.activate
    def test_render_issue_without_beta_comment(self):
        issue_number = self.issue_number_without_beta_comment
        tag = self.tag_beta
        self.mock_util.mock_get_repo()
        # Mock the issue
        api_url = "{}/issues/{}".format(self.repo_api_url, issue_number)
        expected_issue = self._get_expected_issue(issue_number)
        responses.add(method=responses.GET, url=api_url, json=expected_issue)

        # Mock the comments list
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        expected_comment_1 = self._get_expected_issue_comment("Some other comment")
        responses.add(
            method=responses.GET, url=api_url, json=[], content_type="application/json"
        )

        # Mock the comment post response
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        expected_comment_1 = self._get_expected_issue_comment(
            "{} {}".format(
                GithubIssuesParser.ISSUE_COMMENT["beta"], self.version_number_beta
            )
        )
        responses.add(method=responses.POST, url=api_url, json=expected_comment_1)

        generator = self._create_generator(tag)
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": issue_number,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        expected_render = self._create_expected_render(
            issue_number, expected_issue["title"], False
        )
        self.assertEqual(parser.render(), expected_render)
        self.assertEqual(len(responses.calls._calls), 4)

    @responses.activate
    def test_render_issue_with_prod_comment(self):
        issue_number = self.issue_number_with_prod_comment
        tag = self.tag_prod

        self.mock_util.mock_get_repo()

        # Mock the issue
        api_url = "{}/issues/{}".format(self.repo_api_url, issue_number)
        expected_issue = self._get_expected_issue(issue_number)
        responses.add(method=responses.GET, url=api_url, json=expected_issue)

        # Mock the comments list
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        expected_comment_1 = self._get_expected_issue_comment(
            GithubIssuesParser.ISSUE_COMMENT["prod"]
        )
        expected_comments = [expected_comment_1]
        responses.add(method=responses.GET, url=api_url, json=expected_comments)

        generator = self._create_generator(tag)
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": issue_number,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        expected_render = self._create_expected_render(
            issue_number, expected_issue["title"], False
        )
        self.assertEqual(parser.render(), expected_render)
        self.assertEqual(len(responses.calls._calls), 3)

    @responses.activate
    def test_render_issue_without_prod_comment(self):
        issue_number = self.issue_number_without_prod_comment
        tag = self.tag_prod
        self.mock_util.mock_get_repo()
        # Mock the issue
        api_url = "{}/issues/{}".format(self.repo_api_url, issue_number)
        expected_issue = self._get_expected_issue(issue_number)
        responses.add(method=responses.GET, url=api_url, json=expected_issue)

        # Mock the comments list
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        expected_comment_1 = self._get_expected_issue_comment("Some other comment")
        responses.add(
            method=responses.GET, url=api_url, json=[], content_type="application/json"
        )

        # Mock the comment post response
        api_url = "{}/issues/{}/comments".format(self.repo_api_url, issue_number)
        expected_comment_1 = self._get_expected_issue_comment(
            "{} {}".format(
                GithubIssuesParser.ISSUE_COMMENT["prod"], self.version_number_prod
            )
        )
        responses.add(method=responses.POST, url=api_url, json=expected_comment_1)

        generator = self._create_generator(tag)
        parser = GithubIssuesParser(generator, self.title)
        parser.content = [
            {
                "issue_number": issue_number,
                "pr_number": self.pr_number,
                "pr_url": self.pr_url,
            }
        ]
        expected_render = self._create_expected_render(
            issue_number, expected_issue["title"], False
        )
        render = parser.render()
        self.assertEqual(render, expected_render)
        self.assertEqual(len(responses.calls._calls), 4)

    def _create_expected_render(self, issue_number, issue_title, link_pr):
        render = "# {}\r\n\r\n#{}: {}".format(self.title, issue_number, issue_title)
        if link_pr:
            render += " [[PR{}]({})]".format(self.pr_number, self.pr_url)
        return render


class TestInstallLinkParser:
    def test_no_package_version(self):
        generator = mock.Mock(link_pr=True)
        generator.version_id = None
        parser = InstallLinkParser(generator, "Title")
        parser.parse("abc")
        assert parser.render() == ""

    def test_package_version(self):
        generator = mock.Mock(link_pr=True)
        generator.version_id = "foo bar"
        parser = InstallLinkParser(generator, "Title")
        parser.parse("abc")
        output = parser.render()
        assert (
            "https://login.salesforce.com/packaging/installPackage.apexp?p0=foo+bar"
            in output
        )
        assert (
            "https://test.salesforce.com/packaging/installPackage.apexp?p0=foo+bar"
            in output
        )



import json
import os
import tempfile
import unittest
from unittest import mock
from pathlib import Path

import pytest

from cumulusci.core.config import BaseConfig
from cumulusci.core.config import UniversalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import ConnectedAppOAuthConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.keychain import BaseEncryptedProjectKeychain
from cumulusci.core.keychain import EncryptedFileProjectKeychain
from cumulusci.core.keychain import EnvironmentProjectKeychain
from cumulusci.core.keychain.encrypted_file_project_keychain import GlobalOrg
from cumulusci.core.exceptions import ConfigError
from cumulusci.core.exceptions import KeychainKeyNotFound
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import ServiceNotValid
from cumulusci.core.exceptions import OrgNotFound
from cumulusci.core.tests.utils import EnvironmentVarGuard

__location__ = os.path.dirname(os.path.realpath(__file__))


class ProjectKeychainTestMixin(unittest.TestCase):

    keychain_class = BaseProjectKeychain

    def setUp(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"no_yaml": True}
        )
        self.project_config.config["services"] = {
            "connected_app": {"attributes": {"test": {"required": True}}},
            "github": {"attributes": {"name": {"required": True}, "password": {}}},
            "not_configured": {"attributes": {"foo": {"required": True}}},
        }
        self.project_config.project__name = "TestProject"
        self.services = {
            "connected_app": ServiceConfig({"test": "value"}),
            "github": ServiceConfig({"name": "hub"}),
        }
        self.org_config = OrgConfig({"foo": "bar"}, "test")
        self.scratch_org_config = ScratchOrgConfig(
            {"foo": "bar", "scratch": True}, "test_scratch"
        )
        self.key = "0123456789123456"

    def test_init(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(keychain.project_config, self.project_config)
        self.assertEqual(keychain.key, self.key)

    def test_set_non_existant_service(self, project=False):
        keychain = self.keychain_class(self.project_config, self.key)
        with self.assertRaises(ServiceNotValid):
            keychain.set_service("doesnotexist", ServiceConfig({"name": ""}), project)

    def test_set_invalid_service(self, project=False):
        keychain = self.keychain_class(self.project_config, self.key)
        with self.assertRaises(ServiceNotValid):
            keychain.set_service("github", ServiceConfig({"name": ""}), project)

    def test_get_service_not_configured(self):
        keychain = self.keychain_class(self.project_config, self.key)
        with self.assertRaises(ServiceNotConfigured):
            keychain.get_service("not_configured")

    def test_change_key(self):
        new_key = "9876543210987654"
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(self.org_config)
        keychain.set_service("connected_app", self.services["connected_app"])
        keychain.set_service("github", self.services["github"])
        keychain.change_key(new_key)
        self.assertEqual(keychain.key, new_key)
        self.assertEqual(
            keychain.get_service("connected_app").config,
            self.services["connected_app"].config,
        )
        self.assertEqual(
            keychain.get_service("github").config, self.services["github"].config
        )
        self.assertEqual(keychain.get_org("test").config, self.org_config.config)

    def test_set_service_github(self, project=False):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_service("github", self.services["github"], project)
        self.assertEqual(
            keychain.get_service("github").config, self.services["github"].config
        )

    def test_set_and_get_org(self, global_org=False):
        keychain = self.keychain_class(self.project_config, self.key)
        self.org_config.global_org = global_org
        keychain.set_org(self.org_config, global_org)
        self.assertEqual(list(keychain.orgs.keys()), ["test"])
        self.assertEqual(keychain.get_org("test").config, self.org_config.config)

    def test_set_and_get_scratch_org(self, global_org=False):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(self.scratch_org_config, global_org)
        self.assertEqual(list(keychain.orgs.keys()), ["test_scratch"])
        org = keychain.get_org("test_scratch")
        self.assertEqual(org.config, self.scratch_org_config.config)
        self.assertEqual(org.__class__, ScratchOrgConfig)

    def test_load_scratch_orgs_none(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(list(keychain.orgs), [])

    def test_load_scratch_orgs_create_one(self):
        self.project_config.config["orgs"] = {}
        self.project_config.config["orgs"]["scratch"] = {}
        self.project_config.config["orgs"]["scratch"]["test_scratch_auto"] = {}
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(list(keychain.orgs), ["test_scratch_auto"])

    def test_load_scratch_orgs_existing_org(self):
        self.project_config.config["orgs"] = {}
        self.project_config.config["orgs"]["scratch"] = {}
        self.project_config.config["orgs"]["scratch"]["test"] = {}
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(OrgConfig({}, "test"))
        self.assertEqual(list(keychain.orgs), ["test"])
        org = keychain.get_org("test")
        self.assertEqual(org.scratch, None)

    def test_get_org_not_found(self):
        keychain = self.keychain_class(self.project_config, self.key)
        with self.assertRaises(OrgNotFound):
            keychain.get_org("test")

    def test_get_default_org(self):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = self.org_config.config.copy()
        org_config = OrgConfig(org_config, "test", keychain=keychain)
        org_config.save()
        keychain.set_default_org("test")
        org_config.config["default"] = True
        self.assertEqual(keychain.get_default_org()[1].config, org_config.config)

    def test_get_default_org_no_default(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(keychain.get_default_org()[1], None)

    @mock.patch("sarge.Command")
    def test_set_default_org(self, Command):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = self.org_config.config.copy()
        org_config = OrgConfig(org_config, "test")
        keychain.set_org(org_config)
        keychain.set_default_org("test")
        expected_org_config = org_config.config.copy()
        expected_org_config["default"] = True

        self.assertEqual(expected_org_config, keychain.get_default_org()[1].config)

    @mock.patch("sarge.Command")
    def test_unset_default_org(self, Command):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = self.org_config.config.copy()
        org_config = OrgConfig(org_config, "test")
        org_config.config["default"] = True
        keychain.set_org(org_config)
        keychain.unset_default_org()
        self.assertEqual(keychain.get_default_org()[1], None)

    def test_list_orgs(self):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(self.org_config)
        self.assertEqual(keychain.list_orgs(), ["test"])

    def test_list_orgs_empty(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(keychain.list_orgs(), [])


class TestBaseProjectKeychain(ProjectKeychainTestMixin):
    def test_convert_connected_app(self):
        project_config = BaseProjectConfig(
            self.universal_config,
            {
                "services": {
                    "connected_app": {
                        "attributes": {
                            "callback_url": {},
                            "client_id": {},
                            "client_secret": {},
                        }
                    }
                }
            },
        )
        keychain = self.keychain_class(project_config, self.key)
        app_config = {
            "callback_url": "http://localhost:8080/callback",
            "client_id": "CLIENT",
            "client_secret": "SECRET",
        }
        keychain.config["app"] = BaseConfig(app_config)
        keychain._convert_connected_app()
        self.assertEqual(app_config, keychain.get_service("connected_app").config)

    def test_create_scratch_org(self):
        project_config = BaseProjectConfig(
            self.universal_config, {"orgs": {"scratch": {"dev": {}}}}
        )
        keychain = self.keychain_class(project_config, self.key)
        keychain.set_org = mock.Mock()
        keychain.create_scratch_org("test", "dev", days=3)
        org_config = keychain.set_org.call_args[0][0]
        self.assertEqual(3, org_config.days)

    @mock.patch("cumulusci.core.keychain.base_project_keychain.cleanup_org_cache_dirs")
    def test_remove_org(self, cleanup_org_cache_dirs):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(self.org_config)
        keychain.remove_org("test")
        self.assertNotIn("test", keychain.orgs)
        assert cleanup_org_cache_dirs.called_once_with(keychain, self.project_config)


class TestEnvironmentProjectKeychain(ProjectKeychainTestMixin):
    keychain_class = EnvironmentProjectKeychain

    def setUp(self):
        super(TestEnvironmentProjectKeychain, self).setUp()
        self.env = EnvironmentVarGuard().__enter__()
        self._clean_env(self.env)
        self.env.set(
            f"{self.keychain_class.org_var_prefix}test",
            json.dumps(self.org_config.config),
        )
        self.env.set(
            f"{self.keychain_class.service_var_prefix}connected_app",
            json.dumps(self.services["connected_app"].config),
        )
        self.env.set(
            f"{self.keychain_class.service_var_prefix}github",
            json.dumps(self.services["github"].config),
        )

    def tearDown(self):
        self.env.__exit__()

    def _clean_env(self, env):
        for key, value in list(env.items()):
            if key.startswith(self.keychain_class.org_var_prefix):
                del env[key]
        for key, value in list(env.items()):
            if key.startswith(self.keychain_class.service_var_prefix):
                del env[key]

    def test_load_app(self):
        self.env["CUMULUSCI_CONNECTED_APP"] = "{}"
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertIsInstance(keychain.app, ConnectedAppOAuthConfig)

    def test_get_org(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(list(keychain.orgs.keys()), ["test"])
        self.assertEqual(keychain.get_org("test").config, self.org_config.config)

    def test_get_org_not_found(self):
        self._clean_env(self.env)
        super(TestEnvironmentProjectKeychain, self).test_get_org_not_found()

    def test_list_orgs(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(keychain.list_orgs(), ["test"])

    def test_list_orgs_empty(self):
        self._clean_env(self.env)
        self.env.set(
            f"{self.keychain_class.service_var_prefix}connected_app",
            json.dumps(self.services["connected_app"].config),
        )
        super(TestEnvironmentProjectKeychain, self).test_list_orgs_empty()

    def test_load_scratch_org_config(self):
        self._clean_env(self.env)
        self.env.set(
            f"{self.keychain_class.org_var_prefix}test",
            json.dumps(self.scratch_org_config.config),
        )
        keychain = self.keychain_class(self.project_config, self.key)
        self.assertEqual(keychain.list_orgs(), ["test"])
        self.assertEqual(keychain.orgs["test"].__class__, ScratchOrgConfig)

    def test_load_scratch_orgs_create_one(self):
        self._clean_env(self.env)
        super(TestEnvironmentProjectKeychain, self).test_load_scratch_orgs_create_one()

    def test_load_scratch_orgs_none(self):
        self._clean_env(self.env)
        super(TestEnvironmentProjectKeychain, self).test_load_scratch_orgs_none()

    def test_get_default_org(self):
        org_config = self.org_config.config.copy()
        org_config["default"] = True
        self.env.set(
            f"{self.keychain_class.org_var_prefix}test", json.dumps(org_config)
        )
        super(TestEnvironmentProjectKeychain, self).test_get_default_org()

    def test_set_default_org(self):
        """ The EnvironmentProjectKeychain does not persist default org settings """
        org_config = self.org_config.config.copy()
        self.env.set(
            f"{self.keychain_class.org_var_prefix}test", json.dumps(org_config)
        )
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_default_org("test")
        expected_org_config = self.org_config.config.copy()
        expected_org_config["default"] = True

        self.assertEqual(None, keychain.get_default_org()[1])

    def test_set_and_get_scratch_org(self):
        self._clean_env(self.env)
        super(TestEnvironmentProjectKeychain, self).test_set_and_get_scratch_org()


class TestBaseEncryptedProjectKeychain(ProjectKeychainTestMixin):
    keychain_class = BaseEncryptedProjectKeychain

    def test_get_connected_app(self):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.app = keychain._encrypt_config(BaseConfig({}))
        app = keychain.get_connected_app()
        self.assertIsInstance(app, ConnectedAppOAuthConfig)

    def test_decrypt_config__no_config(self):
        keychain = self.keychain_class(self.project_config, self.key)
        config = keychain._decrypt_config(OrgConfig, None, extra=["test", keychain])
        self.assertEqual(config.__class__, OrgConfig)
        self.assertEqual(config.config, {})
        self.assertEqual(config.keychain, keychain)

    def test_decrypt_config__no_config_2(self):
        keychain = self.keychain_class(self.project_config, self.key)
        config = keychain._decrypt_config(BaseConfig, None)
        self.assertEqual(config.__class__, BaseConfig)
        self.assertEqual(config.config, {})

    def test_decrypt_config__wrong_key(self):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(self.org_config, False)

        keychain.key = "x" * 16
        with pytest.raises(KeychainKeyNotFound):
            keychain.get_org("test")

    # def test_decrypt_config__py2_bytes(self):
    #     keychain = self.keychain_class(self.project_config, self.key)
    #     s =
    #     config = keychain._decrypt_config(BaseConfig, s)
    #     assert config["tést"] == "ünicode"

    def test_validate_key__not_set(self):
        with self.assertRaises(KeychainKeyNotFound):
            self.keychain_class(self.project_config, None)

    def test_validate_key__wrong_length(self):
        with self.assertRaises(ConfigError):
            self.keychain_class(self.project_config, "1")


class TestEncryptedFileProjectKeychain(ProjectKeychainTestMixin):
    keychain_class = EncryptedFileProjectKeychain

    def setUp(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.project_config.config["services"] = {
            "connected_app": {"attributes": {"test": {"required": True}}},
            "github": {"attributes": {"git": {"required": True}, "password": {}}},
            "not_configured": {"attributes": {"foo": {"required": True}}},
        }
        self.project_config.project__name = "TestProject"
        self.project_name = "TestProject"
        self.org_config = OrgConfig({"foo": "bar"}, "test")
        self.scratch_org_config = ScratchOrgConfig(
            {"foo": "bar", "scratch": True}, "test_scratch"
        )
        self.services = {
            "connected_app": ServiceConfig({"test": "value"}),
            "github": ServiceConfig({"git": "hub"}),
        }
        self.key = "0123456789123456"

        self._mk_temp_home()
        self._home_patch = mock.patch(
            "pathlib.Path.home", return_value=Path(self.tempdir_home)
        )
        self._home_patch.__enter__()
        self._mk_temp_project()
        os.chdir(self.tempdir_project)

    def tearDown(self):
        self._home_patch.__exit__(None, None, None)

    def _mk_temp_home(self):
        self.tempdir_home = tempfile.mkdtemp()
        global_config_dir = os.path.join(self.tempdir_home, ".cumulusci")
        os.makedirs(global_config_dir)

    def _mk_temp_project(self):
        self.tempdir_project = tempfile.mkdtemp()
        git_dir = os.path.join(self.tempdir_project, ".git")
        os.makedirs(git_dir)
        self._create_git_config()

    def _create_git_config(self):
        filename = os.path.join(self.tempdir_project, ".git", "config")
        content = (
            '[remote "origin"]\n'
            + f"  url = git@github.com:TestOwner/{self.project_name}"
        )
        self._write_file(filename, content)

    def _write_file(self, filename, content):
        with open(filename, "w") as f:
            f.write(content)

    def test_set_service_github_project(self):
        self.test_set_service_github(True)

    def test_set_and_get_org_global(self):
        self.test_set_and_get_org(True)

    def test_set_and_get_org__universal_config(self):
        keychain = self.keychain_class(self.universal_config, self.key)
        keychain.set_org(self.org_config, False)
        self.assertEqual(list(keychain.orgs.keys()), [])

    def test_load_files__empty(self):
        dummy_keychain = BaseEncryptedProjectKeychain(self.project_config, self.key)
        os.makedirs(os.path.join(self.tempdir_home, ".cumulusci", self.project_name))
        self._write_file(
            os.path.join(self.tempdir_home, "test.org"),
            dummy_keychain._encrypt_config(BaseConfig({"foo": "bar"})).decode("utf-8"),
        )
        keychain = self.keychain_class(self.project_config, self.key)
        del keychain.config["orgs"]
        with mock.patch.object(
            self.keychain_class, "global_config_dir", Path(self.tempdir_home)
        ):
            keychain._load_orgs()
        self.assertIn("foo", keychain.get_org("test").config)
        self.assertEqual(keychain.get_org("test").keychain, keychain)

    def test_load_file(self):
        self._write_file(os.path.join(self.tempdir_home, "config"), "foo")
        keychain = self.keychain_class(self.project_config, self.key)
        keychain._load_file(self.tempdir_home, "config", "from_file")
        self.assertEqual("foo", keychain.config["from_file"])

    def test_load_file__universal_config(self):
        self._write_file(os.path.join(self.tempdir_home, "config"), "foo")
        keychain = self.keychain_class(self.project_config, self.key)
        keychain._load_file(self.tempdir_home, "config", "from_file")
        self.assertEqual("foo", keychain.config["from_file"])

    @mock.patch("cumulusci.core.utils.cleanup_org_cache_dirs")
    def test_remove_org(self, cleanup_org_cache_dirs):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.set_org(self.org_config)
        keychain.remove_org("test")
        self.assertNotIn("test", keychain.orgs)
        assert cleanup_org_cache_dirs.called_once_with(keychain, self.project_config)

    def test_remove_org__not_found(self):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.orgs["test"] = mock.Mock()
        with self.assertRaises(OrgNotFound):
            keychain.remove_org("test")

    def test_remove_org__global__not_found(self):
        keychain = self.keychain_class(self.project_config, self.key)
        keychain.orgs["test"] = mock.Mock()
        with self.assertRaises(OrgNotFound):
            keychain.remove_org("test", global_org=True)

    def test_set_and_get_org_local_should_not_shadow_global(self):
        keychain = self.keychain_class(self.project_config, self.key)
        self.org_config.global_org = True
        keychain.set_org(self.org_config, global_org=True)
        assert ["test"] == list(keychain.orgs.keys())
        assert isinstance(keychain.orgs["test"], GlobalOrg), keychain.orgs["test"]
        assert self.org_config.config == keychain.get_org("test").config
        assert Path(self.tempdir_home, ".cumulusci", "test.org").exists()

        # check that it saves to the right place
        with mock.patch(
            "cumulusci.core.keychain.encrypted_file_project_keychain.open"
        ) as o:
            self.org_config.save()
            opened_filename = o.mock_calls[0][1][0]
            assert ".cumulusci/test.org" in opened_filename.replace(
                os.sep, "/"
            ), opened_filename

        # check that it can be loaded in a fresh keychain
        new_keychain = self.keychain_class(self.project_config, self.key)
        org_config = new_keychain.get_org("test")
        assert org_config.global_org

    def test_cache_dir(self):
        keychain = self.keychain_class(self.project_config, self.key)
        assert keychain.cache_dir.name == ".cci"

    def test_get_default_org__with_files(self):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = OrgConfig(self.org_config.config.copy(), "test", keychain=keychain)
        org_config.save()
        with open(self._default_org_path(), "w") as f:
            f.write("test")
        try:
            self.assertEqual(keychain.get_default_org()[1].config, org_config.config)
        finally:
            self._default_org_path().unlink()

    def test_get_default_org__with_files__missing_org(self):
        keychain = self.keychain_class(self.project_config, self.key)
        with open(self._default_org_path(), "w") as f:
            f.write("should_not_exist")
        assert self._default_org_path().exists()
        assert keychain.get_default_org() == (None, None)
        assert not self._default_org_path().exists()

    @mock.patch("sarge.Command")
    def test_set_default_org__with_files(self, Command):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = OrgConfig(self.org_config.config.copy(), "test")
        keychain.set_org(org_config)
        keychain.set_default_org("test")
        with open(self._default_org_path()) as f:
            assert f.read() == "test"
        self._default_org_path().unlink()

    @mock.patch("sarge.Command")
    def test_unset_default_org__with_files(self, Command):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = self.org_config.config.copy()
        org_config = OrgConfig(org_config, "test")
        keychain.set_org(org_config)
        keychain.set_default_org("test")
        keychain.unset_default_org()
        self.assertEqual(keychain.get_default_org()[1], None)
        assert not self._default_org_path().exists()

    def _default_org_path(self):
        return Path(self.tempdir_home) / ".cumulusci/TestProject/DEFAULT_ORG.txt"

    # old way of finding defaults used contents of the files themselves
    # we should preserve backwards compatibiliity for a few months
    def test_get_default_org__file_missing_fallback(self):
        keychain = self.keychain_class(self.project_config, self.key)
        org_config = OrgConfig(self.org_config.config.copy(), "test", keychain=keychain)
        org_config.config["default"] = True
        org_config.save()
        self.assertEqual(keychain.get_default_org()[1].config, org_config.config)

    def test_get_default_org__outside_project(self):
        keychain = self.keychain_class(self.universal_config, self.key)
        assert keychain.get_default_org() == (None, None)



# Note: Unlike in traditional robot libraries, this docstring will end
# up in output from the libdoc task. For that to happen, it must be
# before the imports.

"""
The following page objects are automatically included when
importing the PageObject library. You should not directly import this
file.

"""

import re
from SeleniumLibrary.errors import ElementNotFound
from selenium.common.exceptions import NoSuchElementException
from cumulusci.robotframework.pageobjects import pageobject
from cumulusci.robotframework.pageobjects import BasePage
from cumulusci.robotframework.utils import capture_screenshot_on_error

# This will appear in the generated documentation in place of
# the filename.
TITLE = "Base Page Objects"


@pageobject(page_type="Listing")
class ListingPage(BasePage):
    """Page object representing a Listing page

    When going to the Listing page, you need to specify the object name. You
    may also specify the name of a filter.

    Example

    | Go to page  Listing  Contact  filterName=Recent

    """

    def _go_to_page(self, filter_name=None):
        url_template = "{root}/lightning/o/{object_name}/list"
        url = url_template.format(
            root=self.cumulusci.org.lightning_base_url, object_name=self.object_name
        )
        if filter_name:
            url += "?filterName={}".format(filter_name)
        self.selenium.go_to(url)
        self.salesforce.wait_until_loading_is_complete()

    def _is_current_page(self):
        self.selenium.location_should_contain(
            "/lightning/o/{}/list".format(self.object_name)
        )


class ModalMixin:
    def _wait_to_appear(self, expected_heading=None):
        """Waits until the modal is visible"""
        locator = "//div[contains(@class, 'uiModal')]"
        if expected_heading:
            locator += f"//h2[text()='{expected_heading}']"
            error = f"A modal with the heading {expected_heading} did not appear before the timeout"
        else:
            error = "The modal did not appear before the timeout"

        self.salesforce.wait_for_aura()
        self.selenium.wait_until_element_is_visible(locator, error=error)

    @capture_screenshot_on_error
    def close_the_modal(self):
        """ Closes the open modal """

        locator = "css: button.slds-modal__close"
        self.selenium.wait_until_element_is_enabled(locator)
        self.selenium.click_element(locator)
        self.wait_until_modal_is_closed()
        self._remove_from_library_search_order()

    @capture_screenshot_on_error
    def click_modal_button(self, button_label):
        """Click the named modal button (Save, Save & New, Cancel, etc)"""
        # stolen from Salesforce.py:click_modal_button
        locator = (
            "//div[contains(@class,'uiModal')]"
            "//div[contains(@class,'modal-footer') or contains(@class, 'actionsContainer')]"
            "//button[.//span[text()='{}']]"
        ).format(button_label)

        self.selenium.wait_until_page_contains_element(locator)
        self.selenium.wait_until_element_is_enabled(locator)
        self.salesforce._jsclick(locator)

    @capture_screenshot_on_error
    def modal_should_contain_errors(self, *messages):
        """Verify that the modal contains the following errors

        This will look for the given message in the standard SLDS
        component (<ul class='errorsList'>)
        """
        for message in messages:
            locator = "//ul[@class='errorsList']//li[contains(., \"{}\")]".format(
                message
            )
            self.selenium.page_should_contain_element(
                locator,
                'The page did not contain an error with the text "{}"'.format(message),
            )

    @capture_screenshot_on_error
    def populate_field(self, name, value):
        """Populate a field on the modal form

        Name must the the label of a field as it appears on the modal form.

        Example

        | Populate field  First Name  Connor
        | Populate field  Last Name   MacLeod
        """
        # For now, call the same keyword in Salesforce.py. Eventually
        # that keyword may get moved here and deprecated from that
        # library.
        self.salesforce.populate_field(name, value)

    @capture_screenshot_on_error
    def populate_form(self, *args, **kwargs):
        """Populate the modal form

        Arguments are of the form key=value, where 'key' represents
        a field name as it appears on the form (specifically, the text
        of a label with the class 'uiLabel').

        Example:

        | Populate form
        | ...  First Name=Connor
        | ...  Last Name=MacLeod
        """
        # For now, call the same keyword in Salesforce.py. Eventually
        # that keyword may get moved here and deprecated from that
        # library.
        self.salesforce.populate_form(*args, **kwargs)

    @capture_screenshot_on_error
    def select_dropdown_value(self, label, value):
        """Sets the value of a dropdown form field from one of the values in the dropdown

        ``label`` represents the label on the form field (eg: Lead Source)

        If a modal is present, the modal will be searched for the dropdown. Otherwise
        the first matching dropdown on the entire web page will be used.
        """
        input_element = self._get_form_element_for_label(label)
        if input_element is None:
            raise Exception(f"No form element found with the label '{label}'")

        # SeleniumLibrary's scroll_element_into_view doesn't seem to work
        # for elements in a scrollable div on Firefox. Javascript seems to
        # work though.
        self.selenium.driver.execute_script(
            "arguments[0].scrollIntoView()", input_element
        )
        trigger = input_element.find_element_by_class_name("uiPopupTrigger")
        trigger.click()

        try:
            popup_locator = (
                "//div[contains(@class, 'uiPopupTarget')][contains(@class, 'visible')]"
            )
            popup = self.selenium.get_webelement(popup_locator)
        except ElementNotFound:
            raise ElementNotFound("Timed out waiting for the dropdown menu")

        try:
            value_element = popup.find_element_by_link_text(value)
            value_element.click()
        except NoSuchElementException:
            raise Exception(f"Dropdown value '{value}' not found")

    @capture_screenshot_on_error
    def wait_until_modal_is_closed(self, timeout=None):
        """Waits until the modal is no longer visible

        If the modal isn't open, this will not throw an error.
        """
        locator = "//div[contains(@class, 'uiModal')]"

        self.selenium.wait_until_page_does_not_contain_element(locator)

    def _get_form_element_for_label(self, label):
        """Returns an element of class 'slds-form-element' which contains the given label.

        If you use this to find an item on a modal, you're responsible for waiting for the
        modal to appear before calling this function.
        """
        # search from the root of the document, unless a modal
        # is open
        root = self.selenium.driver.find_element_by_tag_name("body")
        with self._no_implicit_wait():
            # We don't want to wait, we just want to know if the
            # modal is alredy there or not. The caller should have
            # already waited for the modal.

            # also, we use find_elements (plural) here because it will
            # return an empty list rather than throwing an error
            modals = root.find_elements_by_xpath("//div[contains(@class, 'uiModal')]")
            # There should only ever be zero or one, but the Salesforce
            # UI never ceases to surprise me.
            modals = [m for m in modals if m.is_displayed()]
            if modals:
                root = modals[0]
        try:
            # gnarly, but effective.
            # this finds an element with the class slds-form-element which
            # has a child element with the class 'form-element__label' and
            # text that matches the given label.
            locator = f".//*[contains(@class, 'slds-form-element') and .//*[contains(@class, 'form-element__label')]/descendant-or-self::text()='{label}']"
            element = root.find_element_by_xpath(locator)
            return element

        except NoSuchElementException:
            raise ElementNotFound(f"Form element with label '{label}' was not found")


@pageobject("New")
class NewModal(ModalMixin, BasePage):
    """A page object representing the New Object modal

    Note: You should not use this page object with 'Go to page'. Instead,
    you can use 'Wait for modal to appear' after performing an action
    that causes the new object modal to appear (eg: clicking the
    "New" button). Once the modal appears, the keywords for that
    modal will be available for use in the test.

    Example:

    | Go to page                 Home  Contact
    | Click object button        New
    | Wait for modal to appear   New  Contact
    """


@pageobject("Edit")
class EditModal(ModalMixin, BasePage):
    """A page object representing the Edit Object modal

    Note: You should not use this page object with 'Go to page'. Instead,
    you can use 'Wait for modal to appear' after performing an action
    that causes the new object modal to appear (eg: clicking the
    "Edit" button). Once the modal appears, the keywords for that
    modal will be available for use in the test.

    Example:

    | Click object button        Edit
    | Wait for modal to appear   Edit  Contact

    """


@pageobject("Home")
class HomePage(BasePage):
    """A page object representing the home page of an object.

    When going to the Home page, you need to specify the object name.

    Note: The home page of an object may automatically redirect you to
    some other page, such as a Listing page. If you are working with
    such a page, you might need to use `page should be` to load the
    keywords for the page you expect to be redirected to.

    Example

    | Go to page  Home  Contact

    """

    def _go_to_page(self):
        url_template = "{root}/lightning/o/{object_name}/home"
        url = url_template.format(
            root=self.cumulusci.org.lightning_base_url, object_name=self.object_name
        )
        self.selenium.go_to(url)
        self.salesforce.wait_until_loading_is_complete()

    def _is_current_page(self):
        self.selenium.location_should_contain(
            "/lightning/o/{}/home".format(self.object_name)
        )


@pageobject("Detail")
class DetailPage(BasePage):
    """A page object representing the standard Detail page.

    When going to this page via the standard `Go to page` keyword, you
    can specify either an object id, or a set of keyword arguments which
    will be used to look up the object id. When using keyword arguments,
    they need to represent a unique user.

    Example

    | ${contact_id} =       Salesforce Insert  Contact
    | ...                   FirstName=${first_name}
    | ...                   LastName=${last_name}
    |
    | # Using the id:
    | Go to page  Detail  Contact  ${contact_id}
    |
    | # Using lookup parameters
    | Go to page  Detail  Contact  FirstName=${first_name}  LastName=${last_name}

    """

    def _go_to_page(self, object_id=None, **kwargs):
        """Go to the detail page for the given record.

        You may pass in an object id, or you may pass in keyword arguments
        which can be used to look up the object.

        Example

        | Go to page  Detail  Contact  firstName=John  lastName=Smith
        """

        if kwargs and object_id:
            raise Exception("Specify an object id or keyword arguments, but not both")

        if kwargs:
            # note: this will raise an exception if no object is found,
            # or if multiple objects are found.
            object_id = self._get_object(**kwargs)["Id"]

        url_template = "{root}/lightning/r/{object_name}/{object_id}/view"
        url = url_template.format(
            root=self.cumulusci.org.lightning_base_url,
            object_name=self.object_name,
            object_id=object_id,
        )
        self.selenium.go_to(url)
        self.salesforce.wait_until_loading_is_complete()

    def _is_current_page(self, **kwargs):
        """Verify we are on a detail page.

        If keyword arguments are present, this function will go a query
        on the given parameters, assert that the query returns a single
        result, and the verify that the returned object id is part of the url.
        """
        if kwargs:
            # do a lookup to get the object id
            object_id = self._get_object(**kwargs)["Id"]
            pattern = r"/lightning/r/{}/{}/view$".format(self.object_name, object_id)
        else:
            # no kwargs means we should just verify we are on a detail
            # page without regard to which object
            pattern = r"/lightning/r/{}/.*/view$".format(self.object_name)

        location = self.selenium.get_location()
        if not re.search(pattern, location):
            raise Exception(
                "Location '{}' didn't match pattern {}".format(location, pattern)
            )



from cumulusci.core.exceptions import DependencyLookupError
import re

import github3

from cumulusci.tasks.github.base import BaseGithubTask
from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask

VERSION_ID_RE = re.compile(r"version_id: (\S+)")


class GetPackageDataFromCommitStatus(BaseGithubTask, BaseSalesforceApiTask):

    task_options = {
        "context": {
            "description": "Name of the commit status context",
            "required": True,
        },
        "version_id": {"description": "Package version id"},
    }

    def _run_task(self):
        repo = self.get_repo()
        context = self.options["context"]
        commit_sha = self.project_config.repo_commit

        dependencies = []
        version_id = self.options.get("version_id")
        if version_id is None:
            try:
                commit = repo.commit(commit_sha)
            except github3.exceptions.NotFoundError:
                raise DependencyLookupError(
                    f"Could not find commit {commit_sha} on github"
                )
            for status in commit.status().statuses:
                if status.state == "success" and status.context == context:
                    match = VERSION_ID_RE.search(status.description)
                    if match:
                        version_id = match.group(1)

        if version_id:
            dependencies = self._get_dependencies(version_id)
        else:
            raise DependencyLookupError(
                f"Could not find package version id in '{context}' commit status."
            )

        self.return_values = {"dependencies": dependencies, "version_id": version_id}

    def _get_dependencies(self, version_id):
        res = self.tooling.query(
            f"SELECT Dependencies FROM SubscriberPackageVersion WHERE Id='{version_id}'"
        )
        if res["records"]:
            subscriber_version = res["records"][0]
            dependencies = subscriber_version["Dependencies"] or {"ids": []}
            dependencies = [
                {"version_id": d["subscriberPackageVersionId"]}
                for d in dependencies["ids"]
            ]
            return dependencies
        else:
            raise DependencyLookupError(
                f"Could not look up dependencies of {version_id}"
            )



from typing import Optional

from cumulusci.core.utils import process_list_arg
from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils.xml.metadata_tree import MetadataElement


class AddRelatedLists(MetadataSingleEntityTransformTask):
    entity = "Layout"
    task_options = {
        "related_list": {
            "description": "Name of the Related List to include",
            "required": True,
        },
        "fields": {
            "description": "Array of field API names to include in the related list",
            "required": False,
        },
        "exclude_buttons": {
            "description": "Array of button names to suppress from the related list"
        },
        "custom_buttons": {
            "description": "Array of button names to add to the related list"
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> Optional[MetadataElement]:
        related_list = self._inject_namespace(self.options["related_list"])
        existing_related_lists = metadata.findall(
            "relatedLists", relatedList=related_list
        )

        if existing_related_lists:
            return None

        self._create_new_related_list(metadata, api_name, related_list)

        return metadata

    def _create_new_related_list(self, metadata, api_name, related_list):
        self.logger.info(f"Adding Related List {related_list} to {api_name}")

        fields = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("fields", []))
        ]
        exclude_buttons = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("exclude_buttons", []))
        ]
        custom_buttons = [
            self._inject_namespace(f)
            for f in process_list_arg(self.options.get("custom_buttons", []))
        ]

        elem = metadata.append("relatedLists")

        for f in fields:
            elem.append("fields", text=f)

        for button in exclude_buttons:
            elem.append("excludeButtons", button)

        for button in custom_buttons:
            elem.append("customButtons", button)

        elem.append("relatedList", text=related_list)



import os
import re
import sys
import contextlib
import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path


@contextlib.contextmanager
def tee_stdout_stderr(args, logger, tempfile):
    """Tee stdout and stderr so that they're also routed to
    a log file. Add the current command arguments
    as the first item in the log."""
    real_stdout_write = sys.stdout.write
    real_stderr_write = sys.stderr.write

    # Add current command args as first line in logfile
    logger.debug(" ".join(args) + "\n")

    def stdout_write(s):
        output = strip_ansi_sequences(s)
        logger.debug(output)
        real_stdout_write(s)

    def stderr_write(s):
        output = strip_ansi_sequences(s)
        logger.debug(output)
        real_stderr_write(s)

    sys.stdout.write = stdout_write
    sys.stderr.write = stderr_write
    try:
        yield
    finally:
        # reset write functions
        sys.stdout.write = real_stdout_write
        sys.stderr.write = real_stderr_write

        # close temporary logfile
        logger.handlers[0].close()

        # log contents of tempfile to rotating log files
        with open(tempfile, "r", encoding="utf-8") as f:
            contents = f.read()

        logger = get_gist_logger()
        logger.debug(contents)
        # delete temporary log file
        os.remove(tempfile)


def get_gist_logger():
    """Determines the appropriate filepath for logfile
    and name for the logger. Returns a logger with
    RotatingFileHandler attached."""
    logfile_dir = Path.home() / ".cumulusci" / "logs"
    logfile_dir.mkdir(parents=True, exist_ok=True)
    logfile_path = logfile_dir / "cci.log"

    return get_rot_file_logger("stdout/stderr", logfile_path)


def get_rot_file_logger(name, path):
    """Returns a logger with a rotating file handler"""
    logger = logging.getLogger(name)

    handler = RotatingFileHandler(path, backupCount=5, encoding="utf-8")
    handler.doRollover()  # rollover existing log files
    handler.terminator = ""  # click.echo already adds a newline
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return logger


def strip_ansi_sequences(input):
    """Strip ANSI sequences from what's in buffer"""
    ansi_escape = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")
    return ansi_escape.sub("", input)



import unittest
from robot.libraries.BuiltIn import RobotNotRunningError
from cumulusci.robotframework.Salesforce import Salesforce
from unittest import mock


# FIXME: we shouldn't have to tweak these tests for every
# version. The tests should be smarter.
class TestLocators(unittest.TestCase):
    @mock.patch("cumulusci.robotframework.Salesforce.Salesforce.get_latest_api_version")
    def test_locators_in_robot_context(self, get_latest_api_version):
        """Verify we can get locators for the current org api version"""
        get_latest_api_version.return_value = 50.0

        # This instantiates the robot library, mimicking a robot library import.
        # We've mocked out the code that would otherwise throw an error since
        # we're not running in the context of a robot test. The library should
        # return the latest version of the locators.
        sf = Salesforce()

        expected = "cumulusci.robotframework.locators_50"
        actual = sf.locators_module.__name__
        message = "expected to load '{}', actually loaded '{}'".format(expected, actual)
        self.assertEqual(expected, actual, message)

        pass

    @mock.patch(
        "robot.libraries.BuiltIn.BuiltIn.get_library_instance",
        side_effect=RobotNotRunningError(),
    )
    def test_locators_outside_robot_context(self, builtin_mock):
        """Verify that we get the latest locators if not running in the context of a robot test"""

        # This instantiates the robot library, mimicing a robot library import
        # however, because we've mocked get_library_instance to throw an error,
        # we expect the library to still be instantiated, but with the latest
        # version of the locators.
        sf = Salesforce()
        expected = "cumulusci.robotframework.locators_50"
        actual = sf.locators_module.__name__
        message = "expected to load '{}', actually loaded '{}'".format(expected, actual)
        self.assertEqual(expected, actual, message)

    def test_locators_50(self):
        """Verify that locators_50 is a superset of the locators_49

        This test is far from perfect, but it should at least flag a
        catastrophic error in how locators for a version that augments
        the locators from previous versions.

        Note: this test assumes that locators_49 doesn't delete any of the
        keys from 49.

        """
        import cumulusci.robotframework.locators_49 as locators_49
        import cumulusci.robotframework.locators_50 as locators_50

        keys_49 = set(locators_49.lex_locators)
        keys_50 = set(locators_50.lex_locators)

        self.assertNotEqual(
            id(locators_49.lex_locators),
            id(locators_50.lex_locators),
            "locators_49.lex_locators and locators_50.lex_locators are the same object",
        )
        self.assertTrue(len(keys_49) > 0)
        self.assertTrue(keys_49.issubset(keys_50))



import unittest
import yaml
from unittest.mock import Mock, patch, call, mock_open

from cumulusci.tasks.salesforce import LoadCustomSettings
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.core.exceptions import TaskOptionsError, CumulusCIException


class TestLoadCustomSettings(unittest.TestCase):
    @patch("os.path.isfile")
    def test_init_options__exception(self, isfile):
        isfile.return_value = False
        with self.assertRaises(TaskOptionsError):
            create_task(LoadCustomSettings, {"settings_path": "test.yml"})

    @patch("os.path.isfile")
    def test_load_settings_bad_yaml(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()

        task.settings = {"Test__c": "Test"}
        with self.assertRaises(CumulusCIException):
            task._load_settings()

    @patch("os.path.isfile")
    def test_load_settings_list_setting(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()

        task.settings = {"Test__c": {"Name": {"Field__c": "Test"}}}
        task._load_settings()

        task.sf.Test__c.upsert.assert_called_once_with(
            "Name/Name", {"Field__c": "Test"}
        )

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_setting_profile(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 0},
        ]
        task.settings = {
            "Test__c": [{"location": {"profile": "Test"}, "data": {"Field__c": "Test"}}]
        }
        task._load_settings()

        task.sf.Test__c.create.assert_called_once_with(
            {"Field__c": "Test", "SetupOwnerId": "001000000000000"}
        )
        task.sf.query.assert_has_calls(
            [
                call("SELECT Id FROM Profile WHERE Name = 'Test'"),
                call("SELECT Id FROM Test__c WHERE SetupOwnerId = '001000000000000'"),
            ]
        )

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_setting_profile__error(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {
                "totalSize": 2,
                "records": [{"Id": "001000000000000"}, {"Id": "001000000000001"}],
            },
            {"totalSize": 0},
        ]
        task.settings = {
            "Test__c": [{"location": {"profile": "Test"}, "data": {"Field__c": "Test"}}]
        }
        with self.assertRaises(CumulusCIException):
            task._load_settings()

        task.sf.Test__c.create.assert_not_called()
        task.sf.query.assert_has_calls(
            [call("SELECT Id FROM Profile WHERE Name = 'Test'")]
        )

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_setting_user(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 0},
        ]
        task.settings = {
            "Test__c": [
                {
                    "location": {"user": {"name": "test@example.com"}},
                    "data": {"Field__c": "Test"},
                }
            ]
        }
        task._load_settings()

        task.sf.Test__c.create.assert_called_once_with(
            {"Field__c": "Test", "SetupOwnerId": "001000000000000"}
        )

        task.sf.query.assert_has_calls(
            [
                call("SELECT Id FROM User WHERE Username = 'test@example.com'"),
                call("SELECT Id FROM Test__c WHERE SetupOwnerId = '001000000000000'"),
            ]
        )

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_setting_user_email(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 0},
        ]
        task.settings = {
            "Test__c": [
                {
                    "location": {"user": {"email": "test-user@example.com"}},
                    "data": {"Field__c": "Test"},
                }
            ]
        }
        task._load_settings()

        task.sf.Test__c.create.assert_called_once_with(
            {"Field__c": "Test", "SetupOwnerId": "001000000000000"}
        )

        task.sf.query.assert_has_calls(
            [
                call("SELECT Id FROM User WHERE Email = 'test-user@example.com'"),
                call("SELECT Id FROM Test__c WHERE SetupOwnerId = '001000000000000'"),
            ]
        )

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_org(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 0},
        ]
        task.settings = {"Test__c": [{"location": "org", "data": {"Field__c": "Test"}}]}
        task._load_settings()

        task.sf.Test__c.create.assert_called_once_with(
            {"Field__c": "Test", "SetupOwnerId": "001000000000000"}
        )

        task.sf.query.assert_has_calls(
            [
                call("SELECT Id FROM Organization"),
                call("SELECT Id FROM Test__c WHERE SetupOwnerId = '001000000000000'"),
            ]
        )

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_no_location(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 0},
        ]
        task.settings = {"Test__c": [{"data": {"Field__c": "Test"}}]}
        with self.assertRaises(CumulusCIException):
            task._load_settings()

        task.sf.Test__c.create.assert_not_called()

    @patch("os.path.isfile")
    def test_load_settings_hierarchy_update(self, isfile):
        isfile.return_value = True
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        task.sf = Mock()
        task.sf.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 1, "records": [{"Id": "001000000000001"}]},
        ]
        task.settings = {"Test__c": [{"location": "org", "data": {"Field__c": "Test"}}]}
        task._load_settings()

        task.sf.Test__c.update.assert_called_once_with(
            "001000000000001", {"Field__c": "Test", "SetupOwnerId": "001000000000000"}
        )

        task.sf.query.assert_has_calls(
            [
                call("SELECT Id FROM Organization"),
                call("SELECT Id FROM Test__c WHERE SetupOwnerId = '001000000000000'"),
            ]
        )

    @patch("os.path.isfile")
    @patch("simple_salesforce.Salesforce")
    def test_run_task(self, sf, isfile):
        isfile.return_value = True
        m = mock_open(
            read_data=yaml.dump(
                {"Test__c": [{"location": "org", "data": {"Field__c": "Test"}}]}
            )
        )
        task = create_task(LoadCustomSettings, {"settings_path": "test.yml"})

        sf.return_value.query.side_effect = [
            {"totalSize": 1, "records": [{"Id": "001000000000000"}]},
            {"totalSize": 1, "records": [{"Id": "001000000000001"}]},
        ]
        with patch("builtins.open", m):
            task()

        task.sf.Test__c.update.assert_called_once_with(
            "001000000000001", {"Field__c": "Test", "SetupOwnerId": "001000000000000"}
        )

        task.sf.query.assert_has_calls(
            [
                call("SELECT Id FROM Organization"),
                call("SELECT Id FROM Test__c WHERE SetupOwnerId = '001000000000000'"),
            ]
        )



import http.client
import responses

from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin


class MockUtil(GithubApiTestMixin):
    BASE_HTML_URL = "https://github.com"
    BASE_API_URL = "https://api.github.com"

    def __init__(self, owner, repo):
        self.owner = owner
        self.repo = repo
        self.init_github()

    @property
    def html_url(self):
        return "{}/{}/{}".format(self.BASE_HTML_URL, self.owner, self.repo)

    @property
    def repo_url(self):
        return "{}/repos/{}/{}".format(self.BASE_API_URL, self.owner, self.repo)

    def mock_edit_release(self, body=None, draft=True, prerelease=False):
        if body is None:
            body = "Test release body"
        responses.add(
            method=responses.PATCH,
            url="{}/releases/1".format(self.repo_url),
            json=self._get_expected_release(
                "1", body=body, draft=draft, prerelease=prerelease
            ),
            status=http.client.OK,
        )

    def mock_get_repo(self):
        responses.add(
            method=responses.GET,
            url=self.repo_url,
            json=self._get_expected_repo(self.owner, self.repo),
            status=http.client.OK,
        )

    def mock_get_release(self, tag, body):
        responses.add(
            method=responses.GET,
            url="{}/releases/tags/{}".format(self.repo_url, tag),
            json=self._get_expected_release(
                tag, url="{}/releases/1".format(self.repo_url), body=body
            ),
            status=http.client.OK,
        )

    def mock_post_comment(self, issue_number):
        responses.add(
            method=responses.POST,
            url="{}/issues/{}/comments".format(self.repo_url, issue_number),
            status=http.client.OK,
        )

    def mock_pull_request(self, pr_number, body, title=None):
        if title is None:
            title = "Test Pull Request Title"
        responses.add(
            method=responses.GET,
            url="{}/pulls/{}".format(self.repo_url, pr_number),
            json=self._get_expected_pull_request(pr_number, pr_number, body=body),
            status=http.client.OK,
        )

    def mock_pulls(
        self, method=responses.GET, pulls=None, head=None, base=None, state=None
    ):
        # Default url params added by github3
        # see github3.repos.repo.py _Repository.pull_requests()
        params = ["sort=created", "direction=desc", "per_page=100"]
        default_num_params = len(params)
        params_added = False

        if head:
            params.append("head={}".format(head))
        if base:
            params.append("base={}".format(base))
        if state:
            params.append("state={}".format(state))

        if len(params) > default_num_params:
            params_added = True
            param_str = "?{}".format("&".join(params))

        api_url = "{}/pulls{}".format(self.repo_url, param_str if params_added else "")
        responses.add(method=method, url=api_url, json=pulls or [])

    def mock_issue(self, issue_num, labels=None, owner=None, repo=None):
        """Args:
        issue_num: int representing number of the issue
        labels: list(str) of labels to associate with the issue
        owner: str ownerof the repo
        repo: str name of the repo"""
        if not labels:
            labels = []

        self.add_issue_response(
            self._get_expected_issue(issue_num, owner=owner, repo=repo, labels=labels)
        )

    def add_issue_response(self, issue_json):
        responses.add(
            method=responses.GET,
            url="{}/issues/{}".format(self.repo_url, issue_json["number"]),
            json=issue_json,
            status=http.client.OK,
        )

    def mock_issue_labels(self, issue_num, method=responses.GET, labels=None):
        if not labels:
            labels = []
        responses.add(
            method=method,
            url="{}/issues/{}/labels".format(self.repo_url, issue_num),
            json=self._get_expected_labels(labels),
            status=http.client.OK,
        )

    def mock_add_labels_to_issue(self, issue_num, labels=None):
        if not labels:
            labels = []

        responses.add(
            method=responses.POST,
            url="{}/issues/{}/labels".format(self.repo_url, issue_num),
            json=self._get_expected_labels(labels),
            status=http.client.OK,
        )

    def mock_pull_request_by_commit_sha(self, commit_sha):
        responses.add(
            method=responses.GET,
            url="{}/commits/{}/pulls".format(self.repo_url, commit_sha),
            json=[self._get_expected_pull_request(1, 1)],
            status=http.client.OK,
        )



import os
from typing import NamedTuple
from pathlib import Path

from cumulusci.core.exceptions import OrgNotFound
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.keychain import BaseEncryptedProjectKeychain
from cumulusci.core.config import OrgConfig


class EncryptedFileProjectKeychain(BaseEncryptedProjectKeychain):
    """ An encrypted project keychain that stores in the project's local directory """

    @property
    def global_config_dir(self):
        try:
            global_config_dir = (
                self.project_config.universal_config_obj.cumulusci_config_dir
            )
        except AttributeError:
            # Handle a global config passed as a project config
            global_config_dir = self.project_config.cumulusci_config_dir
        return global_config_dir

    @property
    def project_local_dir(self):
        return self.project_config.project_local_dir

    def _load_files(self, dirname, extension, key, constructor=None):
        if dirname is None:
            return
        for item in sorted(os.listdir(dirname)):
            if item.endswith(extension):
                with open(os.path.join(dirname, item), "r") as f_item:
                    config = f_item.read()
                name = item.replace(extension, "")
                if key not in self.config:
                    self.config[key] = {}
                self.config[key][name] = constructor(config) if constructor else config

    def _load_file(self, dirname, filename, key):
        if dirname is None:
            return
        full_path = os.path.join(dirname, filename)
        if not os.path.exists(full_path):
            return
        with open(os.path.join(dirname, filename), "r") as f_item:
            config = f_item.read()
        self.config[key] = config

    def _load_app(self):
        self._load_file(self.global_config_dir, "connected.app", "app")
        self._load_file(self.project_local_dir, "connected.app", "app")

    def _load_orgs(self):
        self._load_files(self.global_config_dir, ".org", "orgs", GlobalOrg)

        self._load_files(self.project_local_dir, ".org", "orgs", LocalOrg)

    def _load_services(self):
        self._load_files(self.global_config_dir, ".service", "services")
        self._load_files(self.project_local_dir, ".service", "services")

    def _remove_org(self, name, global_org):
        if global_org:
            full_path = os.path.join(self.global_config_dir, f"{name}.org")
        else:
            full_path = os.path.join(self.project_local_dir, f"{name}.org")
        if not os.path.exists(full_path):
            kwargs = {"name": name}
            if not global_org:
                raise OrgNotFound(
                    "Could not find org named {name} to delete.  Deleting in project org mode.  Is {name} a global org?".format(
                        **kwargs
                    )
                )
            raise OrgNotFound(
                "Could not find org named {name} to delete.  Deleting in global org mode.  Is {name} a project org instead of a global org?".format(
                    **kwargs
                )
            )

        os.remove(full_path)
        del self.orgs[name]

    def _set_encrypted_org(self, name, encrypted, global_org):
        if global_org:
            filename = os.path.join(self.global_config_dir, f"{name}.org")
        elif self.project_local_dir is None:
            return
        else:
            filename = os.path.join(self.project_local_dir, f"{name}.org")
        with open(filename, "wb") as f_org:
            f_org.write(encrypted)

    def _set_encrypted_service(self, name, encrypted, project):
        if project:
            filename = os.path.join(self.project_local_dir, f"{name}.service")
        else:
            filename = os.path.join(self.global_config_dir, f"{name}.service")
        with open(filename, "wb") as f_service:
            f_service.write(encrypted)

    def _raise_org_not_found(self, name):
        raise OrgNotFound(
            f"Org information could not be found. Expected to find encrypted file at {self.project_local_dir}/{name}.org"
        )

    def _raise_service_not_configured(self, name):
        raise ServiceNotConfigured(
            f"'{name}' service configuration could not be found. "
            f"Maybe you need to run: cci service connect {name}"
        )

    def _get_org(self, name):
        org = self._decrypt_config(
            OrgConfig,
            self.orgs[name].encrypted_data,
            extra=[name, self],
            context=f"org config ({name})",
        )
        if self.orgs[name].global_org:
            org.global_org = True
        return org

    @property
    def _default_org_path(self):
        if self.project_local_dir:
            return Path(self.project_local_dir) / "DEFAULT_ORG.txt"

    def get_default_org(self):
        """ Retrieve the name and configuration of the default org """
        # first look for a file with the default org in it
        default_org_path = self._default_org_path
        if default_org_path and default_org_path.exists():
            org_name = default_org_path.read_text().strip()
            try:
                org_config = self.get_org(org_name)
                return org_name, org_config
            except OrgNotFound:  # org was deleted
                default_org_path.unlink()  # we don't really have a usable default anymore

        # fallback to old way of doing it
        org_name, org_config = super().get_default_org()
        if org_name:
            self.set_default_org(org_name)  # upgrade to new way
        return org_name, org_config

    def set_default_org(self, name: str):
        """ Set the default org for tasks and flows by name """
        super().set_default_org(name)
        self._default_org_path.write_text(name)

    def unset_default_org(self):
        """Unset the default orgs for tasks and flows """
        super().unset_default_org()
        try:
            self._default_org_path.unlink()
        except FileNotFoundError:
            pass


class GlobalOrg(NamedTuple):
    encrypted_data: bytes
    global_org: bool = True


class LocalOrg(NamedTuple):
    encrypted_data: bytes
    global_org: bool = False



from cumulusci.robotframework.pageobjects import BasePage
from cumulusci.robotframework.pageobjects import pageobject


@pageobject(page_type="Test", object_name="Foo__c")
class FooTestPage(BasePage):
    def foo_keyword_1(self, message):
        self.builtin.log(message)
        return "foo keyword 1: {}".format(message)



from cumulusci.core.config.SfdxOrgConfig import SfdxOrgConfig
from typing import Optional
import base64
import enum
import io
import json
import pathlib
import zipfile

from pydantic import BaseModel, validator
from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.exceptions import DependencyLookupError, ServiceNotConfigured
from cumulusci.core.exceptions import PackageUploadFailure
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_bool_arg
from cumulusci.salesforce_api.package_zip import BasePackageZipBuilder
from cumulusci.salesforce_api.package_zip import MetadataPackageZipBuilder
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.core.sfdx import get_default_devhub_username
from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask
from cumulusci.tasks.salesforce.org_settings import build_settings_package
from cumulusci.utils import download_extract_github


class PackageTypeEnum(str, enum.Enum):
    managed = "Managed"
    unlocked = "Unlocked"


class VersionTypeEnum(str, enum.Enum):
    major = "major"
    minor = "minor"
    patch = "patch"


class PackageConfig(BaseModel):
    package_name: str
    description: str = ""
    package_type: PackageTypeEnum
    org_dependent: bool = False
    namespace: Optional[str]
    branch: Optional[str] = None
    version_name: str
    version_type: VersionTypeEnum = VersionTypeEnum.minor

    @validator("org_dependent")
    def org_dependent_must_be_unlocked(cls, v, values):
        if v and values["package_type"] != PackageTypeEnum.unlocked:
            raise ValueError("Only unlocked packages can be org-dependent.")
        return v


class CreatePackageVersion(BaseSalesforceApiTask):
    """Creates a new second-generation package version.

    If a package named ``package_name`` does not yet exist in the Dev Hub, it will be created.
    """

    api_version = "49.0"

    task_options = {
        "package_name": {"description": "Name of package"},
        "package_type": {
            "description": "Package type (Unlocked or Managed)",
            "required": True,
        },
        "namespace": {"description": "Package namespace"},
        "version_name": {"description": "Version name"},
        "version_type": {
            "description": "The part of the version number to increment. "
            "Options are major, minor, patch.  Defaults to minor"
        },
        "skip_validation": {
            "description": "If true, skip validation of the package version. Default: false. "
            "Skipping validation creates packages more quickly, but they cannot be promoted for release."
        },
        "org_dependent": {
            "description": "If true, create an org-dependent unlocked package. Default: false."
        },
        "force_upload": {
            "description": "If true, force creating a new package version even if one with the same contents already exists"
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        self.package_config = PackageConfig(
            package_name=self.options.get("package_name")
            or self.project_config.project__package__name,
            package_type=self.options.get("package_type")
            or self.project_config.project__package__type,
            org_dependent=self.options.get("org_dependent", False),
            namespace=self.options.get("namespace")
            or self.project_config.project__package__namespace,
            branch=self.project_config.repo_branch,
            version_name=self.options.get("version_name") or "Release",
            version_type=self.options.get("version_type") or "minor",
        )
        self.options["skip_validation"] = process_bool_arg(
            self.options.get("skip_validation", False)
        )
        self.options["force_upload"] = process_bool_arg(
            self.options.get("force_upload", False)
        )

    def _init_task(self):
        self.devhub_config = self._init_devhub()
        self.tooling = get_simple_salesforce_connection(
            self.project_config,
            self.devhub_config,
            api_version=self.api_version,
            base_url="tooling",
        )

    def _init_devhub(self):
        # Determine the devhub username for this project
        try:
            devhub_service = self.project_config.keychain.get_service("devhub")
        except ServiceNotConfigured:
            devhub_username = get_default_devhub_username()
        else:
            devhub_username = devhub_service.username
        return SfdxOrgConfig({"username": devhub_username}, "devhub")

    def _run_task(self):
        """Creates a new 2GP package version.

        1. Create package if not found in Dev Hub.
        2. Request creation of package version.
        3. Wait for completion.
        4. Collect package information as return values.
        """
        # find existing package in Dev Hub, or create one if necessary
        self.package_id = self._get_or_create_package(self.package_config)
        self.return_values["package_id"] = self.package_id

        # submit request to create package version
        package_zip_builder = MetadataPackageZipBuilder(
            path=self.project_config.default_package_path,
            name=self.package_config.package_name,
            options={"package_type": self.package_config.package_type.value},
            logger=self.logger,
        )
        self.request_id = self._create_version_request(
            self.package_id,
            self.package_config,
            package_zip_builder,
            self.options["skip_validation"],
        )
        self.return_values["request_id"] = self.request_id

        # wait for request to complete
        self._poll()
        self.return_values["package2_version_id"] = self.package_version_id

        # get the new version number from Package2Version
        res = self.tooling.query(
            "SELECT MajorVersion, MinorVersion, PatchVersion, BuildNumber, SubscriberPackageVersionId FROM Package2Version WHERE Id='{}' ".format(
                self.package_version_id
            )
        )
        package2_version = res["records"][0]
        self.return_values["subscriber_package_version_id"] = package2_version[
            "SubscriberPackageVersionId"
        ]
        self.return_values["version_number"] = self._get_version_number(
            package2_version
        )

        # get the new version's dependencies from SubscriberPackageVersion
        res = self.tooling.query(
            "SELECT Dependencies FROM SubscriberPackageVersion "
            f"WHERE Id='{package2_version['SubscriberPackageVersionId']}'"
        )
        self.return_values["dependencies"] = res["records"][0]["Dependencies"]

        self.logger.info("Created package version:")
        self.logger.info(f"  Package2 Id: {self.package_id}")
        self.logger.info(f"  Package2Version Id: {self.package_version_id}")
        self.logger.info(
            f"  SubscriberPackageVersion Id: {self.return_values['subscriber_package_version_id']}"
        )
        self.logger.info(f"  Version Number: {self.return_values['version_number']}")
        self.logger.info(f"  Dependencies: {self.return_values['dependencies']}")

    def _get_or_create_package(self, package_config: PackageConfig):
        """Find or create the Package2

        Checks the Dev Hub for an existing, non-deprecated 2GP package
        with matching name, type, and namespace.
        """
        message = f"Checking for existing {package_config.package_type} Package named {package_config.package_name}"
        query = (
            f"SELECT Id, ContainerOptions FROM Package2 WHERE IsDeprecated = FALSE "
            f"AND ContainerOptions='{package_config.package_type}' "
            f"AND IsOrgDependent={package_config.org_dependent} "
            f"AND Name='{package_config.package_name}'"
        )
        if package_config.namespace:
            query += f" AND NamespacePrefix='{package_config.namespace}'"
            message += f" with namespace {package_config.namespace}"
        else:
            query += " AND NamespacePrefix=null"
        self.logger.info(message)
        try:
            res = self.tooling.query(query)
        except SalesforceMalformedRequest as err:
            if "Object type 'Package2' is not supported" in err.content[0]["message"]:
                raise TaskOptionsError(
                    "This org does not have a Dev Hub with 2nd-generation packaging enabled. "
                    "Make sure you are using the correct org and/or check the Dev Hub settings in Setup."
                )
            raise  # pragma: no cover
        if res["size"] > 1:
            raise TaskOptionsError(
                f"Found {res['size']} packages with the same name, namespace, and package_type"
            )
        if res["size"] == 1:
            existing_package = res["records"][0]
            if existing_package["ContainerOptions"] != package_config.package_type:
                raise PackageUploadFailure(
                    f"Duplicate Package: {existing_package['ContainerOptions']} package with id "
                    f"{ existing_package['Id']} has the same name ({package_config.package_name}) "
                    "for this namespace but has a different package type"
                )
            package_id = existing_package["Id"]
            self.logger.info(f"Found {package_id}")
            return package_id

        self.logger.info("No existing package found, creating the package")
        Package2 = self._get_tooling_object("Package2")
        package = Package2.create(
            {
                "ContainerOptions": package_config.package_type,
                "IsOrgDependent": package_config.org_dependent,
                "Name": package_config.package_name,
                "Description": package_config.description,
                "NamespacePrefix": package_config.namespace,
            }
        )
        return package["id"]

    def _create_version_request(
        self,
        package_id: str,
        package_config: PackageConfig,
        package_zip_builder: BasePackageZipBuilder,
        skip_validation: bool = False,
        dependencies: list = None,
    ):
        # Prepare the VersionInfo file
        version_bytes = io.BytesIO()
        version_info = zipfile.ZipFile(version_bytes, "w", zipfile.ZIP_DEFLATED)
        try:

            # Add the package.zip
            package_hash = package_zip_builder.as_hash()
            version_info.writestr("package.zip", package_zip_builder.as_bytes())

            if not self.options["force_upload"]:
                # Check for an existing package with the same contents
                res = self.tooling.query(
                    "SELECT Id "
                    "FROM Package2VersionCreateRequest "
                    f"WHERE Package2Id = '{package_id}' "
                    "AND Status != 'Error' "
                    f"AND SkipValidation = {str(skip_validation)} "
                    f"AND Tag = 'hash:{package_hash}' "
                    "ORDER BY CreatedDate DESC"
                )
                if res["size"] > 0:
                    self.logger.info(
                        "Found existing request for package with the same metadata.  Using existing package."
                    )
                    return res["records"][0]["Id"]

            # Create the package descriptor
            # @@@ we should support releasing a successor to an older version by specifying a base version
            last_version_parts = self._get_highest_version_parts(package_id)
            version_number = self._get_next_version_number(
                last_version_parts, package_config.version_type
            )
            package_descriptor = {
                "ancestorId": "",  # @@@ need to add this for Managed 2gp
                "id": package_id,
                "path": "",
                "versionName": package_config.version_name,
                "versionNumber": version_number,
            }

            # Add org shape
            with open(self.org_config.config_file, "r") as f:
                scratch_org_def = json.load(f)
            for key in (
                "country",
                "edition",
                "language",
                "features",
                "snapshot",
            ):
                if key in scratch_org_def:
                    package_descriptor[key] = scratch_org_def[key]

            # Add settings
            if "settings" in scratch_org_def:
                with build_settings_package(
                    scratch_org_def["settings"], self.api_version
                ) as path:
                    settings_zip_builder = MetadataPackageZipBuilder(path=path)
                    version_info.writestr(
                        "settings.zip", settings_zip_builder.as_bytes()
                    )

            # Add the dependencies for the package
            is_dependency = package_config is not self.package_config
            if not package_config.org_dependent and not is_dependency:
                self.logger.info("Determining dependencies for package")
                dependencies = self._get_dependencies()
            if dependencies:
                package_descriptor["dependencies"] = dependencies

            # Add package descriptor to version info
            version_info.writestr(
                "package2-descriptor.json", json.dumps(package_descriptor)
            )
        finally:
            version_info.close()
        version_info = base64.b64encode(version_bytes.getvalue()).decode("utf-8")
        Package2CreateVersionRequest = self._get_tooling_object(
            "Package2VersionCreateRequest"
        )
        request = {
            "Branch": package_config.branch,
            "Package2Id": package_id,
            "SkipValidation": skip_validation,
            "Tag": f"hash:{package_hash}",
            "VersionInfo": version_info,
        }
        self.logger.info(
            f"Requesting creation of package version {version_number} "
            f"for package {package_config.package_name} ({package_id})"
        )
        response = Package2CreateVersionRequest.create(request)
        self.logger.info(
            f"Package2VersionCreateRequest created with id {response['id']}"
        )
        return response["id"]

    def _get_highest_version_parts(self, package_id):
        """Get the version parts for the highest existing version of the specified package."""
        res = self.tooling.query(
            "SELECT MajorVersion, MinorVersion, PatchVersion, BuildNumber, IsReleased "
            "FROM Package2Version "
            f"WHERE Package2Id='{package_id}' "
            "ORDER BY MajorVersion DESC, MinorVersion DESC, PatchVersion DESC, BuildNumber DESC "
            "LIMIT 1"
        )
        if res["size"]:
            return res["records"][0]
        return {
            "MajorVersion": 0,
            "MinorVersion": 0,
            "PatchVersion": 0,
            "BuildNumber": 0,
            "IsReleased": False,
        }

    def _get_next_version_number(self, version_parts, version_type: VersionTypeEnum):
        """Predict the next package version.

        Given existing version parts (major/minor/patch) and a version type,
        determine the number to request for the next version.
        """
        new_version_parts = {
            "MajorVersion": version_parts["MajorVersion"],
            "MinorVersion": version_parts["MinorVersion"],
            "PatchVersion": version_parts["PatchVersion"],
            "BuildNumber": "NEXT",
        }
        if version_parts["IsReleased"]:
            if version_type == VersionTypeEnum.major:
                new_version_parts["MajorVersion"] += 1
                new_version_parts["MinorVersion"] = 0
                new_version_parts["PatchVersion"] = 0
            if version_type == VersionTypeEnum.minor:
                new_version_parts["MinorVersion"] += 1
                new_version_parts["PatchVersion"] = 0
            elif version_type == VersionTypeEnum.patch:
                new_version_parts["PatchVersion"] += 1
        return self._get_version_number(new_version_parts)

    def _get_version_number(self, version):
        """Format version fields from Package2Version as a version number."""
        return "{MajorVersion}.{MinorVersion}.{PatchVersion}.{BuildNumber}".format(
            **version
        )

    def _get_dependencies(self):
        """Resolve dependencies into SubscriberPackageVersionIds (04t prefix)"""
        dependencies = self.project_config.get_static_dependencies()

        # If any dependencies are expressed as a 1gp namespace + version,
        # we need to convert those to 04t package version ids,
        # for which we need an org with the packages installed.
        if self._has_1gp_namespace_dependency(dependencies):
            dependencies = self.org_config.resolve_04t_dependencies(dependencies)

        # Convert dependencies to correct format for Package2VersionCreateRequest
        dependencies = self._convert_project_dependencies(dependencies)

        # Build additional packages for local unpackaged/pre
        dependencies = self._get_unpackaged_pre_dependencies(dependencies)

        return dependencies

    def _has_1gp_namespace_dependency(self, project_dependencies):
        """Returns true if any dependencies are specified using a namespace rather than 04t"""
        for dependency in project_dependencies:
            if "namespace" in dependency:
                return True
            if "dependencies" in dependency:
                if self._has_1gp_namespace_dependency(dependency["dependencies"]):
                    return True
        return False

    def _convert_project_dependencies(self, dependencies):
        """Convert dependencies into the format expected by Package2VersionCreateRequest.

        For dependencies expressed as a github repo subfolder, build an unlocked package from that.
        """
        new_dependencies = []
        for dependency in dependencies:
            if dependency.get("dependencies"):
                new_dependencies.extend(
                    self._convert_project_dependencies(dependency["dependencies"])
                )

            new_dependency = {}
            if dependency.get("version_id"):
                name = (
                    f"{dependency['namespace']}@{dependency['version']} "
                    if "namespace" in dependency
                    else ""
                )
                self.logger.info(
                    f"Adding dependency {name} with id {dependency['version_id']}"
                )
                new_dependency["subscriberPackageVersionId"] = dependency["version_id"]

            elif dependency.get("repo_name"):
                version_id = self._create_unlocked_package_from_github(
                    dependency, new_dependencies
                )
                self.logger.info(
                    "Adding dependency {}/{} {} with id {}".format(
                        dependency["repo_owner"],
                        dependency["repo_name"],
                        dependency["subfolder"],
                        version_id,
                    )
                )
                new_dependency["subscriberPackageVersionId"] = version_id

            else:
                raise DependencyLookupError(
                    f"Unable to convert dependency: {dependency}"
                )

            new_dependencies.append(new_dependency)

        return new_dependencies

    def _get_unpackaged_pre_dependencies(self, dependencies):
        """Create package for unpackaged/pre metadata, if necessary"""
        path = pathlib.Path("unpackaged", "pre")
        if not path.exists():
            return dependencies

        for item_path in sorted(path.iterdir(), key=str):
            if item_path.is_dir():
                version_id = self._create_unlocked_package_from_local(
                    item_path, dependencies
                )
                self.logger.info(
                    "Adding dependency {}/{} {} with id {}".format(
                        self.project_config.repo_owner,
                        self.project_config.repo_name,
                        item_path,
                        version_id,
                    )
                )
                dependencies.append({"subscriberPackageVersionId": version_id})

        return dependencies

    def _create_unlocked_package_from_github(self, dependency, dependencies):
        gh_for_repo = self.project_config.get_github_api(
            dependency["repo_owner"], dependency["repo_name"]
        )
        zip_src = download_extract_github(
            gh_for_repo,
            dependency["repo_owner"],
            dependency["repo_name"],
            dependency["subfolder"],
            ref=dependency.get("ref"),
        )
        package_zip_builder = MetadataPackageZipBuilder.from_zipfile(
            zip_src, options=dependency, logger=self.logger
        )

        package_config = PackageConfig(
            package_name="{repo_owner}/{repo_name} {subfolder}".format(**dependency),
            version_name="Auto",
            package_type="Unlocked",
            # Ideally we'd do this without a namespace,
            # but it needs to match the dependent package
            namespace=self.package_config.namespace,
        )
        package_id = self._get_or_create_package(package_config)
        self.request_id = self._create_version_request(
            package_id,
            package_config,
            package_zip_builder,
            dependencies=dependencies,
        )

        self._poll()
        self._reset_poll()
        res = self.tooling.query(
            "SELECT SubscriberPackageVersionId FROM Package2Version "
            f"WHERE Id='{self.package_version_id}'"
        )
        package2_version = res["records"][0]
        return package2_version["SubscriberPackageVersionId"]

    def _create_unlocked_package_from_local(self, path, dependencies):
        """Create an unlocked package version from a local directory."""
        self.logger.info("Creating package for dependencies in {}".format(path))
        package_name = (
            f"{self.project_config.repo_owner}/{self.project_config.repo_name} {path}"
        )
        package_zip_builder = MetadataPackageZipBuilder(
            path=path, name=package_name, logger=self.logger
        )
        package_config = PackageConfig(
            package_name=package_name,
            version_name="Auto",
            package_type="Unlocked",
            # Ideally we'd do this without a namespace,
            # but it needs to match the dependent package
            namespace=self.package_config.namespace,
        )
        package_id = self._get_or_create_package(package_config)
        self.request_id = self._create_version_request(
            package_id, package_config, package_zip_builder, dependencies=dependencies
        )
        self._poll()
        self._reset_poll()
        res = self.tooling.query(
            "SELECT SubscriberPackageVersionId FROM Package2Version "
            f"WHERE Id='{self.package_version_id}'"
        )
        package2_version = res["records"][0]
        return package2_version["SubscriberPackageVersionId"]

    def _poll_action(self):
        """Check if Package2VersionCreateRequest has completed."""
        res = self.tooling.query(
            f"SELECT Id, Status, Package2VersionId FROM Package2VersionCreateRequest WHERE Id = '{self.request_id}'"
        )
        request = res["records"][0]
        if request["Status"] == "Success":
            self.logger.info("[Success]: Package creation successful")
            self.poll_complete = True
            self.package_version_id = request["Package2VersionId"]
        elif request["Status"] == "Error":
            self.logger.error("[Error]: Package creation failed with error:")
            res = self.tooling.query(
                "SELECT Message FROM Package2VersionCreateRequestError "
                f"WHERE ParentRequestId = '{request['Id']}'"
            )
            errors = []
            if res["size"] > 0:
                for error in res["records"]:
                    errors.append(error["Message"])
                    self.logger.error(error["Message"])
            raise PackageUploadFailure("\n".join(errors))
        else:
            self.logger.info(f"[{request['Status']}]")
